/*
 * [Generate.java]
 *
 * Summary: To generate database and Java source code files used by MimeCheck.
 *
 * Copyright: (c) 2008-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2008-02-20 Created with IntelliJ IDEA.
 */
package com.mindprod.mimecheck;

import com.mindprod.common18.BigDate;
import com.mindprod.common18.Build;
import com.mindprod.common18.EIO;
import com.mindprod.csv.CSVReader;
import com.mindprod.fastcat.FastCat;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;

import static java.lang.System.*;

interface Selector
    {
    /**
     * @param e an Extension
     *
     * @return true if should include this Extension in the output
     */
    @SuppressWarnings( { "BooleanMethodIsAlwaysInverted" } )
    boolean include( MimeData e );
    }

/**
 * To generate database and Java source code files used by MimeCheck.
 * <p/>
 * Run Generate, then manually insert E:/com/mindprod/mimecheck/include.javafrag into
 * MimeCheck.Java Generate CSV database of MIME info into Java source, HTML for mime.html and HTML for sound.html
 * Database E:/com/mindprod/mimecheck/mimes.csv look like this:
 * <p/>
 * P, extension, mime (for primary mime type )
 * <p/>
 * S, extension, mime (for secondary mime type )
 * <p/>
 * C, extension, comment (for a comment about the extension )
 * <p/>
 * A, extension (this is an audio mime)
 * <p/>
 * V, extension (this is video  * mime)
 * <p/>
 * W, extension (this is core type that should typically be supported by a Website)
 * <p/>
 * May appear in any order
 * We generated HTML without <table> column headers or </tbody></table>
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.0 2008-02-20 Created with IntelliJ IDEA.
 * @since 2008-02-20
 */
public final class Generate
    {
    public static final String DO_NOT_EDIT = "<!-- The following table was generated on " + BigDate.localToday().toString() + " by mimecheck.Generate. D o   n o t   e d i t . -->\n";

    /**
     * file where mime data are stored in CSV format
     */
    private static final String MIMES_CSV_DATABASE = Build.MINDPROD_SOURCE + "/mimecheck/mimes.csv";

    /**
     * RAM resident version of the mimes database
     */
    private static final HashMap<String, MimeData> h = new HashMap<>( 2000 );

    private static boolean REPORT_UPPER_CASE = false;

    /**
     * database of all extensions in order by extension
     */
    private static MimeData[] sortedMimeData;

    /**
     * generate HTML to be included in for typical core website in  mime.html
     *
     * @param outFilename where to put the generated html
     * @param s           selector to describe which records to extract
     *
     * @throws java.io.IOException if trouble writing
     */
    @SuppressWarnings( { "SameParameterValue" } )
    private static void generateCoreWebsiteMimesHtml( String outFilename, Selector s ) throws IOException
        {
        // typical generated line looks like this:
        //  <tr>
        // <td>aac</td>
        // <td>audio/aac</td>
        // </tr>
        final PrintWriter htmlfrag = EIO.getPrintWriter( new File( outFilename ), 2 * 1048, EIO.UTF8 );
        int everyTenthLine = 0;
        htmlfrag.print( DO_NOT_EDIT );
        for ( MimeData e : sortedMimeData )
            {
            if ( !s.include( e ) )
                {
                continue;
                }
            htmlfrag.print( "<tr><td class=\"ext\">" );
            htmlfrag.print( e.getExt() );
            htmlfrag.print( "</td><td class=\"mime\">" );
            htmlfrag.print( e.getPrimaryMimeType() );
            htmlfrag.print( "</td></tr>\n" );
            } // end for
        htmlfrag.close();
        } // end generateCoreWebsiteMimesHtml

    /**
     * generate serialised form of Mime data for inclusion as a resource in the jar.
     *
     * @param outFilename name of file where generated ser file goes.
     *
     * @throws java.io.IOException if trouble writing
     */
    @SuppressWarnings( { "SameParameterValue" } )
    private static void generateMimeDat( String outFilename ) throws IOException
        {
        // We could create a HashMap and serialise it, but that conflicts with generics, so we use a DataOutputStream.
        // mime.dat is a DataOutputStream, not serialised, not Gzipped.
        //   O P E N
        FileOutputStream fos = new FileOutputStream( outFilename, false );
        BufferedOutputStream bos = new BufferedOutputStream( fos, 8 * 1024 /* buffsize in bytes */ );
        DataOutputStream dos = new DataOutputStream( bos );
        // stream has count of entries, then records of string extension,  count of mimes, string[] mime types
        dos.writeInt( sortedMimeData.length );
        for ( MimeData e : sortedMimeData )
            {
            //  W R I T E   one rec
            dos.writeUTF( e.getExt() );
            String[] mimes = e.getMimeTypes();
            dos.writeInt( mimes.length );
            for ( String s : mimes )
                {
                dos.writeUTF( s );
                }
            }
        // C L O S E
        dos.close();
        } // generateMimeDat

    /**
     * generate HTML to be included in mime.html
     *
     * @param outFilename where to put the generated HTML
     * @param s           selector to describe which records to extract
     *
     * @throws java.io.IOException if trouble writing
     */
    private static void generateMimeHtml( String outFilename, Selector s ) throws IOException
        {
        // typical generated line looks like this:
        //  <tr>
        // <td>aac</td>
        // <td>audio/aac</td>
        // <td>audio/x-aac, audio/x-mp4, audio/MP4-LATM</td>
        // <td>Apple QuickTime compressed audio used in iPods</td>
        // </tr>
        final PrintWriter htmlfrag = EIO.getPrintWriter( new File( outFilename ), 2 * 1048, EIO.UTF8 );
        htmlfrag.print( DO_NOT_EDIT );
        for ( MimeData e : sortedMimeData )
            {
            if ( !s.include( e ) )
                {
                continue;
                }
            final String[] secondaries = e.getSecondaryMimeTypes();
            final FastCat sb = new FastCat( 10 + secondaries.length * 2 );
            sb.append( "<tr><td class=\"ext\">" );
            sb.append( e.getExt() );
            sb.append( "</td><td class=\"mime\">" );
            sb.append( e.getPrimaryMimeType() );
            sb.append( "</td><td class=\"mime\">" );
            if ( secondaries.length == 0 )
                {
                sb.append( "&nbsp;" );
                }
            else
                {
                boolean first = true;
                for ( String secondary : secondaries )
                    {
                    if ( first )
                        {
                        first = false;
                        }
                    else
                        {
                        sb.append( ", " );
                        }
                    sb.append( secondary );
                    }
                }
            sb.append( "</td><td>" );
            final String comment = e.getComment();
            sb.append( comment.length() == 0 ? "&nbsp;" : comment );
            sb.append( "</td></tr>\n" );
            htmlfrag.print( sb.toString() );
            } // end for
        htmlfrag.close();
        } // end generateMimeHtml

    /**
     * load the mime database into RAM
     *
     * @throws IOException if can read mime.csv
     */
    private static void loadMimeDatabase()
            throws IOException
        {
        final CSVReader mimeCSV =
                new CSVReader( new BufferedReader( new FileReader( MIMES_CSV_DATABASE ) ) );
        try
            {
            //noinspection InfiniteLoopStatement
            while ( true )
                {
                final String command = mimeCSV.get();
                if ( command == null || command.length() != 1 )
                    {
                    throw new IllegalArgumentException( "bad command [" + command + "]" );
                    }
                switch ( command.charAt( 0 ) )
                    {
                    case 'P':
                    {
                    final String ext = mimeCSV.get();
                    final String mime = mimeCSV.get();
                    if ( mime.indexOf( '/' ) < 0 )
                        {
                        err.println( "Missing slash in mime P " + mime );
                        }
                    MimeData e = h.get( ext );
                    if ( e == null )
                        {
                        e = new MimeData( ext );
                        h.put( ext, e );
                        }
                    if ( e.getPrimaryMimeType().length() > 0 )
                        {
                        err.println( "P " + mime + " duplicate primary for " + ext );
                        }
                    e.setPrimaryMimeType( mime );
                    break;
                    }
                    case 'S':
                    {
                    final String ext = mimeCSV.get();
                    final String mime = mimeCSV.get();
                    if ( mime.indexOf( '/' ) < 0 )
                        {
                        err.println( "Missing slash in mime S " + mime );
                        }
                    MimeData e = h.get( ext );
                    if ( e == null )
                        {
                        e = new MimeData( ext );
                        h.put( ext, e );
                        }
                    if ( mime.equalsIgnoreCase( e.getPrimaryMimeType() ) )
                        {
                        err.println( "secondary S " + mime + "  duplicates primary for " + ext );
                        break;
                        }
                    String[] existingSecondaries = e.getSecondaryMimeTypes();
                    for ( String candidate : existingSecondaries )
                        {
                        if ( mime.equalsIgnoreCase( candidate ) )
                            {
                            {
                            err.println( "secondary S " + mime + " duplicates secondary for " + ext );
                            break;
                            }
                            }
                        }
                    e.addSecondaryMimeType( mime );
                    break;
                    }
                    case 'C':
                    {
                    final String ext = mimeCSV.get();
                    final String comment = mimeCSV.get();
                    MimeData e = h.get( ext );
                    if ( e == null )
                        {
                        e = new MimeData( ext );
                        h.put( ext, e );
                        }
                    if ( e.getComment().length() > 0 )
                        {
                        err.println( "duplicate comment for " + ext );
                        }
                    if ( comment.length() == 0 )
                        {
                        err.println( "empty comment for " + ext );
                        }
                    e.setComment( comment );
                    break;
                    }
                    case 'A':
                    {
                    final String ext = mimeCSV.get();
                    MimeData e = h.get( ext );
                    if ( e == null )
                        {
                        e = new MimeData( ext );
                        h.put( ext, e );
                        }
                    e.markAsAudio();
                    break;
                    }
                    case 'V':
                    {
                    final String ext = mimeCSV.get();
                    MimeData e = h.get( ext );
                    if ( e == null )
                        {
                        e = new MimeData( ext );
                        h.put( ext, e );
                        }
                    e.markAsVideo();
                    break;
                    }
                    case 'W':
                    {
                    final String ext = mimeCSV.get();
                    MimeData e = h.get( ext );
                    if ( e == null )
                        {
                        e = new MimeData( ext );
                        h.put( ext, e );
                        }
                    e.markAsCoreWebsite();
                    break;
                    }
                    default:
                        throw new IllegalArgumentException( "bad command" );
                    }
                mimeCSV.skipToNextLine();
                } // end while
            }
        catch ( EOFException e )
            {
            mimeCSV.close();
            }
        }

    /**
     * sort the in-RAM mime databas by ext.
     */
    private static void sortMimeDatabase()
        {
        sortedMimeData = h.values().toArray( new MimeData[ h.size() ] );
        Arrays.sort( sortedMimeData );
        }

    /**
     * validate the in-RAM mime databaset.
     */
    private static void validateMimeDatabase()
        {
        for ( MimeData e : h.values() )
            {
            String ext = e.getExt();
            if ( REPORT_UPPER_CASE && !ext.toLowerCase().equals( ext ) )
                {
                err.println( "ext should be lower case: " + e.toString() );
                }
            String primaryMime = e.getPrimaryMimeType();
            if ( primaryMime == null || primaryMime.length() == 0 )
                {
                err.println( "missing primary mime type: " + e.toString() );
                continue;
                }
            if ( REPORT_UPPER_CASE && !primaryMime.toLowerCase().equals( primaryMime ) )
                {
                err.println( "warning: primary Mime usually lower case: " + e.toString() );
                }
            for ( String secondaryMime : e.getSecondaryMimeTypes() )
                {
                if ( REPORT_UPPER_CASE && !secondaryMime.toLowerCase().equals( secondaryMime ) )
                    {
                    err.println( "warning: secondary Mime usually lower case: " + secondaryMime );
                    }
                }
            }
        }

    @SuppressWarnings( { "InfiniteLoopStatement" } )
    public static void main( String[] args )
        {
        try
            {
            loadMimeDatabase();
            validateMimeDatabase();
            sortMimeDatabase();
            generateMimeDat( Build.MINDPROD_SOURCE + "/mimecheck/mimes.dat" );
            generateMimeHtml( Build.MINDPROD_WEBROOT + "/jgloss/include/allmimes.htmlfrag", new Selector()
                {
                public boolean include( MimeData e )
                    {
                    return true;
                    }
                } );
            generateMimeHtml( Build.MINDPROD_WEBROOT + "/jgloss/include/audiomimes.htmlfrag", new Selector()
                {
                public boolean include( MimeData e )
                    {
                    return e.isAudio();
                    }
                } );
            generateMimeHtml( Build.MINDPROD_WEBROOT + "/jgloss/include/videomimes.htmlfrag", new Selector()
                {
                public boolean include( MimeData e )
                    {
                    return e.isVideo();
                    }
                } );
            generateCoreWebsiteMimesHtml( Build.MINDPROD_WEBROOT + "/jgloss/include/corewebsitemimes.htmlfrag",
                    new Selector()
                        {
                        public boolean include( MimeData e )
                            {
                            return e.isCoreWebsite();
                            }
                        }
            );
            }
        catch ( IOException e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "I/O error" + e.getMessage() );
            err.println();
            }
        }
    } // end Generate

/**
 * describes the mime types for one extension
 */
class MimeData implements Comparable<MimeData>
    {
    /**
     * possibly empty, sorted list of secondary mimes
     */
    private final ArrayList<String> secondaryMimeTypes = new ArrayList<>( 11 );

    /**
     * extension
     */
    private final String ext;/* usually 3 long */

    /**
     * comment with embedded entities and HTML tags. empty comment encoded as "", not null.
     */
    private String comment = "";

    /**
     * primary mime
     */
    private String primaryMimeType = "";

    /**
     * true if this is an audio mime
     */
    private boolean audio;

    /**
     * true if this is a core website mime
     */
    private boolean coreWebsite;

    /**
     * true if this is a video mime
     */
    private boolean video;

    /**
     * constructor
     *
     * @param ext file extension, usually three letters.
     */
    MimeData( String ext )
        {
        this.ext = ext;
        }

    /**
     * add another secondary mime type
     *
     * @param secondaryMimeType alternate mime type
     */
    public void addSecondaryMimeType( String secondaryMimeType )
        {
        if ( secondaryMimeType == null || secondaryMimeType.length() == 0 )
            {
            throw new IllegalArgumentException( secondaryMimeType + " missing secondary mime type" );
            }
        this.secondaryMimeTypes.add( secondaryMimeType );
        if ( secondaryMimeType.startsWith( "audio/" ) )
            {
            markAsAudio();
            }
        else if ( secondaryMimeType.startsWith( "video/" ) )
            {
            markAsVideo();
            }
        }

    /**
     * Sort by extension.
     * Defines default the sort order for MimeData Objects.
     * Compare this MimeData with another MimeData.
     * Compares ext.
     * Informally, returns (this-other) or +ve if this is more positive than other.
     *
     * @param other other MimeData to compare with this one
     *
     * @return +ve if this&gt;other, 0 if this==other, -ve if this&lt;other
     */
    public final int compareTo( MimeData other )
        {
        return this.ext.compareToIgnoreCase( other.ext );
        }

    /**
     * get comment.  If none returns "", not null.
     *
     * @return comment on extension/mime.
     */
    public String getComment()
        {
        return comment;
        }

    public void setComment( String comment )
        {
        if ( comment == null )
            {
            this.comment = "";
            }
        else
            {
            this.comment = comment;
            }
        }

    public String getExt()
        {
        return ext;
        }

    /**
     * get sorted list of mime types, including both primary and secondary
     *
     * @return possibly empty array of  mime types
     */
    public String[] getMimeTypes()
        {
        int size = secondaryMimeTypes.size() + 1;
        String[] mimeTypes = new String[ size ];
        mimeTypes[ 0 ] = getPrimaryMimeType();
        Collections.sort( secondaryMimeTypes );
        String[] secondaries = secondaryMimeTypes.toArray( new String[ secondaryMimeTypes.size() ] );
        System.arraycopy( secondaries, 0, mimeTypes, 1, secondaries.length );
        return mimeTypes;
        }

    public String getPrimaryMimeType()
        {
        return primaryMimeType;
        }

    public void setPrimaryMimeType( String primaryMimeType )
        {
        if ( primaryMimeType == null || primaryMimeType.length() == 0 )
            {
            throw new IllegalArgumentException( primaryMimeType + " missing primary mime type" );
            }
        this.primaryMimeType = primaryMimeType;
        if ( primaryMimeType.startsWith( "audio/" ) )
            {
            markAsAudio();
            }
        else if ( primaryMimeType.startsWith( "video/" ) )
            {
            markAsVideo();
            }
        }

    /**
     * get sorted list of secondary mime types
     *
     * @return possibly empty array of secondary mime types
     */
    public String[] getSecondaryMimeTypes()
        {
        Collections.sort( secondaryMimeTypes );
        return secondaryMimeTypes.toArray( new String[ secondaryMimeTypes.size() ] );
        }

    /**
     * return true if this is an audio extension
     *
     * @return true if this is an audio type mime.
     */
    public boolean isAudio()
        {
        return audio;
        }

    /**
     * return true if this is a core Website extension
     *
     * @return true if this is a core Website mime
     */
    public boolean isCoreWebsite()
        {
        return coreWebsite;
        }

    /**
     * return true if this is an video extension
     *
     * @return true if this is a video type mime
     */
    public boolean isVideo()
        {
        return video;
        }

    public void markAsAudio()
        {
        this.audio = true;
        }

    public void markAsCoreWebsite()
        {
        this.coreWebsite = true;
        }

    public void markAsVideo()
        {
        this.video = true;
        }

    /**
     * display this entry as a string for debugging
     *
     * @return representation of all fields separated by colons
     */
    public String toString()
        {
        StringBuilder sb = new StringBuilder( 1000 );
        sb.append( ext );
        sb.append( ":" );
        sb.append( primaryMimeType );
        if ( secondaryMimeTypes.size() > 0 )
            {
            sb.append( ":[" );
            for ( String sec : secondaryMimeTypes )
                {
                sb.append( sec );
                sb.append( ":" );
                }
            sb.setLength( sb.length() - 1 );
            sb.append( "]" );
            }
        if ( comment.length() > 0 )
            {
            sb.append( ":" );
            sb.append( comment );
            }
        if ( audio )
            {
            sb.append( ":" );
            sb.append( 'A' );
            }
        if ( video )
            {
            sb.append( ":" );
            sb.append( 'V' );
            }
        if ( coreWebsite )
            {
            sb.append( ":" );
            sb.append( 'W' );
            }
        return sb.toString();
        }
    }
